<?php
/**********************************************************************************
* convert.php                                                                     *
***********************************************************************************
* SMF: Simple Machines Forum                                                      *
* Open-Source Project Inspired by Zef Hemel (zef@zefhemel.com)                    *
* =============================================================================== *
* Software Version:           SMF 1.1                                             *
* Software by:                Simple Machines (http://www.simplemachines.org)     *
* Copyright 2006-2009 by:     Simple Machines LLC (http://www.simplemachines.org) *
*           2001-2006 by:     Lewis Media (http://www.lewismedia.com)             *
* Support, News, Updates at:  http://www.simplemachines.org                       *
v**********************************************************************************
* This program is free software; you may redistribute it and/or modify it under   *
* the terms of the provided license as published by Simple Machines LLC.          *
*                                                                                 *
* This program is distributed in the hope that it is and will be useful, but      *
* WITHOUT ANY WARRANTIES; without even any implied warranty of MERCHANTABILITY    *
* or FITNESS FOR A PARTICULAR PURPOSE.                                            *
*                                                                                 *
* See the "license.txt" file for details of the Simple Machines license.          *
* The latest version can always be found at http://www.simplemachines.org.        *
**********************************************************************************/

$GLOBALS['required_php_version'] = '4.1.0';
$GLOBALS['required_mysql_version'] = '3.23.28';

// Buy some time
set_time_limit(0);

// When in debug mode, we log our errors. Request is not trusted yet.
if (isset($_GET['debug']) || isset($_POST['debug']))
	set_error_handler('convert_error_hanlder');

// We now have CLI support.
if ((php_sapi_name() == 'cli' || (isset($_SERVER['TERM']) && $_SERVER['TERM'] == 'xterm')) && empty($_SERVER['REMOTE_ADDR']))
{
	$command_line = true;

	cmdStep0();
	exit;
}
else
{
	$command_line = false;
	initialize_inputs();
}

show_header();

if (!empty($_GET['step']) && ($_GET['step'] == 1 || $_GET['step'] == 2))
	echo '
			<div class="panel">
				<h2>Converting...</h2>';

if (function_exists('doStep' . $_GET['step']))
	call_user_func('doStep' . $_GET['step']);

if (!empty($_GET['step']) && ($_GET['step'] == 1 || $_GET['step'] == 2))
	echo '
			</div>';

show_footer();

function initialize_inputs()
{
	global $sourcedir;

	// Save here so it doesn't get overwritten when sessions are restarted.
	$convert_script = @$_REQUEST['convert_script'];

	// Clean up after unfriendly php.ini settings.
	if (function_exists('set_magic_quotes_runtime'))
		@set_magic_quotes_runtime(0);
	error_reporting(E_ALL);
	ignore_user_abort(true);
	umask(0);

	define('SMF', 1);
	ob_start();

	if (@ini_get('session.save_handler') == 'user')
		@ini_set('session.save_handler', 'files');
	@session_start();

	// Add slashes, as long as they aren't already being added.
	if (function_exists('get_magic_quotes_gpc') && @get_magic_quotes_gpc() != 0)
		$_POST = stripslashes_recursive($_POST);

	// This is really quite simple; if ?delete is on the URL, delete the converter...
	if (isset($_GET['delete']))
	{
		@unlink(dirname(__FILE__) . '/convert_error_log');
		@unlink(__FILE__);
		if (preg_match('~_to_smf\.(php|sql)$~', $_SESSION['convert_script']) != 0)
			@unlink(dirname(__FILE__) . '/' . $_SESSION['convert_script']);
		$_SESSION['convert_script'] = null;

		exit;
	}
	// Empty the error log?
	if (isset($_REQUEST['empty_error_log']))
		@unlink(dirname(__FILE__) . '/convert_error_log');

	// The current step - starts at 0.
	$_GET['step'] = (int) @$_GET['step'];
	$_REQUEST['start'] = (int) @$_REQUEST['start'];

	// Check for the password...
	if (isset($_POST['db_pass']))
		$_SESSION['convert_db_pass'] = $_POST['db_pass'];
	elseif (isset($_SESSION['convert_db_pass']))
		$_POST['db_pass'] = $_SESSION['convert_db_pass'];

	if (isset($_SESSION['convert_paths']) && !isset($_POST['path_from']) && !isset($_POST['path_to']))
		list ($_POST['path_from'], $_POST['path_to']) = $_SESSION['convert_paths'];
	elseif (isset($_POST['path_from']) || isset($_POST['path_to']))
	{
		if (isset($_POST['path_from']))
			$_POST['path_from'] = substr($_POST['path_from'], -1) == '/' ? substr($_POST['path_from'], 0, -1) : $_POST['path_from'];
		if (isset($_POST['path_to']))
			$_POST['path_to'] = substr($_POST['path_to'], -1) == '/' ? substr($_POST['path_to'], 0, -1) : $_POST['path_to'];

		$_SESSION['convert_paths'] = array(@$_POST['path_from'], @$_POST['path_to']);
	}

	// If we have $convert_script then set it to the session.
	if (!empty($convert_script))
		$_SESSION['convert_script'] = $convert_script;
	if (isset($_SESSION['convert_script']) && file_exists(dirname(__FILE__) . '/' . $_SESSION['convert_script']) && preg_match('~_to_smf\.(php|sql)$~', $_SESSION['convert_script']) != 0)
	{
		if (substr($_SESSION['convert_script'], -4) == '.php')
			preparse_php(dirname(__FILE__) . '/' . $_SESSION['convert_script']);
		else
			preparse_sql(dirname(__FILE__) . '/' . $_SESSION['convert_script']);
	}
	else
		unset($_SESSION['convert_script']);
}

function preparse_sql($script_filename)
{
	global $convert_data;

	$fp = fopen($script_filename, 'r');
	$data = fread($fp, 4096);
	fclose($fp);

	$convert_data['parameters'] = array();

	// This loads name, settings, table_test, from_prefix, defines, and globals.
	preg_match_all('~---\~ ([^:]+):\s*(.+?)\s*\n~', $data, $matches);
	for ($i = 0, $n = count($matches[1]); $i < $n; $i++)
	{
		// String value?
		if (in_array($matches[1][$i], array('name', 'table_test', 'from_prefix', 'version', 'database_type')))
			$convert_data[$matches[1][$i]] = stripslashes(substr(trim($matches[2][$i]), 1, -1));
		// No... so it must be an array.
		else
		{
			if (strpos($matches[2][$i], '"') === false)
				preg_match_all('~()([^,]+?)(,|$)~', trim($matches[2][$i]), $values);
			else
				preg_match_all('~(")?([^,]+?)\\1(,|$)~', trim($matches[2][$i]), $values);

			if (!isset($convert_data[$matches[1][$i]]))
				$convert_data[$matches[1][$i]] = array();
			$convert_data[$matches[1][$i]] = array_merge($convert_data[$matches[1][$i]], $values[2]);
		}
	}

	if (empty($convert_data['defines']))
		$convert_data['defines'] = array();
	if (empty($convert_data['globals']))
		$convert_data['globals'] = array();
	if (empty($convert_data['settings']))
		$convert_data['settings'] = array();
	if (empty($convert_data['variable']))
		$convert_data['variable'] = array();

	if (!empty($convert_data['parameters']))
	{
		foreach ($convert_data['parameters'] as $i => $param)
		{
			if (is_array($param))
				continue;

			list ($id, $label) = explode('=', $param);
			if (strpos($id, ' ') !== false)
				list ($id, $type) = explode(' ', $id);
			else
				$type = 'text';

			$convert_data['parameters'][$i] = array(
				'id' => $id,
				'label' => $label,
				'type' => $type,
			);
			$convert_data['globals'][] = $id;
		}
	}

	foreach ($convert_data['globals'] as $k => $v)
	{
		$v = trim($v);
		$convert_data['globals'][$k] = $v{0} == '$' ? substr($v, 1) : $v;
	}

	if (isset($_POST['path_to']) && !empty($_GET['step']))
		loadSettings();
}

function preparse_php($script_filename)
{
	global $convert_data;

	$preparsing = true;
	require($script_filename);

	if (empty($convert_data['parameters']))
		$convert_data['parameters'] = array();
	if (empty($convert_data['defines']))
		$convert_data['defines'] = array();
	if (empty($convert_data['globals']))
		$convert_data['globals'] = array();
	if (empty($convert_data['settings']))
		$convert_data['settings'] = array();
	if (empty($convert_data['variable']))
		$convert_data['variable'] = array();

	foreach ($convert_data['globals'] as $k => $v)
	{
		$v = trim($v);
		$convert_data['globals'][$k] = $v{0} == '$' ? substr($v, 1) : $v;
	}

	if (isset($_POST['path_to']) && !empty($_GET['step']))
		loadSettings();
}

function loadSettings()
{
	global $convert_data, $from_prefix, $to_prefix, $convert_dbs, $command_line;

	foreach ($convert_data['defines'] as $define)
	{
		$define = explode('=', $define);
		define($define[0], isset($define[1]) ? $define[1] : '1');
	}
	foreach ($convert_data['globals'] as $global)
		global $$global;

	// Cannot find Settings.php?
	if (!$command_line && !file_exists($_POST['path_to'] . '/Settings.php'))
	{
		show_header();
		return doStep0('This converter was unable to find SMF in the path you specified.<br /><br />Please double check the path, and that it is already installed there.');
	}
	elseif ($command_line && !file_exists($_POST['path_to'] . '/Settings.php'))
		return print_error('This converter was unable to find SMF in the path you specified.<br /><br />Please double check the path, and that it is already installed there.', true);

	$found = empty($convert_data['settings']);
	foreach ($convert_data['settings'] as $file)
		$found |= file_exists($_POST['path_from'] . $file);

	/*
		Check if open_basedir is enabled.  If it's enabled and the converter file was not found then that means
		that the user hasn't moved the files to the public html dir.  With this enabled and the file not found, we can't go anywhere from here.
	*/
	if (!$command_line && @ini_get('open_basedir') != '' && !$found)
	{
		show_header();
		return doStep0('The converter detected that your host has open_basedir enabled on this server.  Please ask your host to disable this setting or try moving the contents of your ' . $convert_data['name'] . ' to the public html folder of your site.');
	}
	elseif ($command_line && @ini_get('open_basedir') != '' && !$found)
		return print_error('The converter detected that your host has open_basedir enabled on this server.  Please ask your host to disable this setting or try moving the contents of your ' . $convert_data['name'] . ' to the public html folder of your site.', true);

	if (!$command_line && !$found)
	{
		show_header();
		return doStep0('Unable to find the settings for ' . $convert_data['name'] . '.  Please double check the path and try again.');
	}
	elseif ($command_line && !$found)
		return print_error('Unable to find the settings for ' . $convert_data['name'] . '.  Please double check the path and try again.', true);

	// Any parameters to speak of?
	if (!empty($convert_data['parameters']) && !empty($_SESSION['convert_parameters']))
	{
		foreach ($convert_data['parameters'] as $param)
		{
			if (isset($_POST[$param['id']]))
				$_SESSION['convert_parameters'][$param['id']] = $_POST[$param['id']];
		}

		// Should already be global'd.
		foreach ($_SESSION['convert_parameters'] as $k => $v)
			$$k = $v;
	}
	elseif (!empty($convert_data['parameters']))
	{
		$_SESSION['convert_parameters'] = array();
		foreach ($convert_data['parameters'] as $param)
		{
			if (isset($_POST[$param['id']]))
				$_SESSION['convert_parameters'][$param['id']] = $_POST[$param['id']];
			else
				$_SESSION['convert_parameters'][$param['id']] = null;
		}

		foreach ($_SESSION['convert_parameters'] as $k => $v)
			$$k = $v;
	}

	// Everything should be alright now... no cross server includes, we hope...
	require_once($_POST['path_to'] . '/Settings.php');
	$GLOBALS['boardurl'] = $boardurl;

	if (!$command_line && $_SESSION['convert_db_pass'] != $db_passwd)
	{
		show_header();
		return doStep0('The database password you entered was incorrect.  Please make sure you are using the right password (for the SMF user!) and try it again.  If in doubt, use the password from Settings.php in the SMF installation.');
	}
	elseif ($command_line && $_SESSION['convert_db_pass'] != $db_passwd)
		return print_error('The database password you entered was incorrect.  Please make sure you are using the right password (for the SMF user!) and try it again.  If in doubt, use the password from Settings.php in the SMF installation.', true);

	// Check the steps that we have decided to go through.
	if (!$command_line && isset($_POST['do_steps']) && empty($_POST['do_steps']))
	{
		show_header();
		return doStep0('You must select at least one step to convert.');
	}
	elseif (!$command_line && isset($_POST['do_steps']))
	{
		unset($_SESSION['do_steps']);
		foreach ($_POST['do_steps'] as $next_step_line => $step)
		{
			//$explode = explode(',', $step);
			//$_SESSION['do_steps'][$key] = array(
			//	'cur_step_line' => $explode[0],
			//	'prev_step_line' => $explode[1],
			//	'next_prev_line' => $explode[2],
			//);
			$_SESSION['do_steps'][$next_step_line] = $step;
		}
	}

	if (isset($_SESSION['convert_parameters']['database_type']) && !isset($convert_data['database_type']))
		$convert_data['database_type'] = $_SESSION['convert_parameters']['database_type'];
	if (isset($convert_data['database_type']) && (function_exists($convert_data['database_type'] . '_query') || function_exists($convert_data['database_type'] . '_exec') || ($convert_data['database_type'] == 'ado' && class_exists('com'))))
	{
		$convert_dbs = $convert_data['database_type'];

		if (isset($convert_data['connect_string']))
			$connect_string = eval('return "' . $convert_data['connect_string'] . '";');
		elseif (isset($_SESSION['convert_parameters']['connect_string']))
			$connect_string = $_SESSION['convert_parameters']['connect_string'];

		if ($convert_dbs == 'odbc')
			$GLOBALS['odbc_connection'] = odbc_connect($connect_string, '', '');
		elseif ($convert_dbs == 'ado')
		{
			$GLOBALS['ado_connection'] = new COM('ADODB.Connection');
			$GLOBALS['ado_connection']->Open($connect_string);

			register_shutdown_function(create_function('', '$GLOBALS[\'ado_connection\']->Close();'));
		}
	}
	elseif (!$command_line && isset($convert_data['database_type']))
	{
		show_header();
		return doStep0('PHP doesn\'t support the database type this converter was written for, \'' . $convert_data['database_type'] . '\'.');
	}
	elseif ($command_line && isset($convert_data['database_type']))
		return print_error('PHP doesn\'t support the database type this converter was written for, \'' . $convert_data['database_type'] . '\'.', true);
	else
		$convert_dbs = 'mysql';

	// Persist?
	if (empty($db_persist))
		mysql_connect($db_server, $db_user, $db_passwd);
	else
		mysql_pconnect($db_server, $db_user, $db_passwd);

	// UTF8
	$charset = findSupportedCharsets();
	$charset = array_flip($charset);
	$charset = isset($_POST['charsets']) && isset($charset[$_POST['charsets']]) ? $_POST['charsets'] : '';
	$charset = !empty($charset) ? $charset : (isset($db_character_set) && preg_match('~^\w+$~', $db_character_set) === 1 ? $db_character_set : '');
	if (!empty($charset))
		mysql_query("SET NAMES $charset");

	if (strpos($db_prefix, '.') === false)
		$to_prefix = is_numeric(substr($db_prefix, 0, 1)) ? $db_name . '.' . $db_prefix : '`' . $db_name . '`.' . $db_prefix;
	else
		$to_prefix = $db_prefix;

	foreach ($convert_data['variable'] as $eval_me)
		eval($eval_me);

	foreach ($convert_data['settings'] as $file)
	{
		if (file_exists($_POST['path_from'] . $file) && empty($convert_data['flatfile']))
			require_once($_POST['path_from'] . $file);
	}

	if (isset($convert_data['from_prefix']))
		$from_prefix = eval('return "' . $convert_data['from_prefix'] . '";');

	if (preg_match('~^`[^`]+`.\d~', $from_prefix) != 0)
		$from_prefix = strtr($from_prefix, array('`' => ''));

	if ($_REQUEST['start'] == 0 && empty($_GET['substep']) && empty($_GET['cstep']) && ($_GET['step'] == 1 || $_GET['step'] == 2) && isset($convert_data['table_test']))
	{
		$result = convert_query("
			SELECT COUNT(*)
			FROM " . eval('return "' . $convert_data['table_test'] . '";'), true);
		if (!$command_line && $result === false)
		{
			show_header();
			doStep0('Sorry, the database connection information used in the specified installation of SMF cannot access the installation of ' . $convert_data['name'] . '.  This may either mean that the installation doesn\'t exist, or that the MySQL account used does not have permissions to access it.<br /><br />The error MySQL gave was: ' . mysql_error());
		}
		elseif ($command_line && $result === false)
			print_error("Sorry, the database connection information used in the specified installation of SMF cannot access the installation of " . $convert_data['name'] . ".  This may either mean that the installation doesn\'t exist, or that the MySQL account used does not have permissions to access it.\n\nThe error MySQL gave was: " . mysql_error(), true);
		convert_free_result($result);
	}
}

function find_convert_scripts()
{
	if (isset($_REQUEST['convert_script']))
	{
		if ($_REQUEST['convert_script'] != '' && preg_match('~^[a-z0-9\-_\.]*_to_smf\.(sql|php)$~i', $_REQUEST['convert_script']) != 0)
		{
			$_SESSION['convert_script'] = preg_replace('~[\.]+~', '.', $_REQUEST['convert_script']);
			if (substr($_SESSION['convert_script'], -4) == '.php')
				preparse_php(dirname(__FILE__) . '/' . $_SESSION['convert_script']);
			else
				preparse_sql(dirname(__FILE__) . '/' . $_SESSION['convert_script']);
		}
		else
			$_SESSION['convert_script'] = null;
	}

	$preparsing = true;

	$dir = dir(dirname(__FILE__));
	$scripts = array();
	while ($entry = $dir->read())
	{
		if (substr($entry, -11) != '_to_smf.sql' && substr($entry, -11) != '_to_smf.php')
			continue;

		$fp = fopen(dirname(__FILE__) . '/' . $entry, 'r');
		$data = fread($fp, 4096);
		fclose($fp);

		if (substr($entry, -11) == '_to_smf.sql')
		{
			if (preg_match('~---\~ name:\s*"(.+?)"~', $data, $match) != 0)
				$scripts[] = array('path' => $entry, 'name' => $match[1]);
		}
		elseif (substr($entry, -11) == '_to_smf.php')
		{
			if (preg_match('~\$convert_data =~', $data) != 0)
			{
				require(dirname(__FILE__) . '/' . $entry);
				$scripts[] = array('path' => $entry, 'name' => $convert_data['name']);
			}
		}
	}
	$dir->close();

	if (isset($_SESSION['convert_script']))
	{
		if (count($scripts) > 1)
			$GLOBALS['possible_scripts'] = $scripts;
		return false;
	}

	if (count($scripts) == 1)
	{
		$_SESSION['convert_script'] = basename($scripts[0]['path']);
		if (substr($_SESSION['convert_script'], -4) == '.php')
			preparse_php(dirname(__FILE__) . '/' . $_SESSION['convert_script']);
		else
			preparse_sql(dirname(__FILE__) . '/' . $_SESSION['convert_script']);
		return false;
	}

	echo '
		<div class="panel">
			<h2>Which software are you using?</h2>';

	if (!empty($scripts))
	{
		echo '
			<h3>The converter found multiple conversion data files.  Please choose the one you wish to use.</h3>

			<ul>';

		foreach ($scripts as $script)
			echo '
				<li><a href="', $_SERVER['PHP_SELF'], '?convert_script=', $script['path'], '">', $script['name'], '</a> <em>(', $script['path'], ')</em></li>';

		echo '
			</ul>

			<h2>It\'s not here!</h2>
			<h3>If the software you\'re looking for doesn\'t appear above, please check to see if it is available for download at <a href="http://www.simplemachines.org/">www.simplemachines.org</a>.  If it isn\'t, we may be able to write one for you - just ask us!</h3>

			If you\'re having any other problems with this converter, don\'t hesitate to look for help on our <a href="http://www.simplemachines.org/community/index.php">forum</a>.';
	}
	else
	{
		echo '
			<h3>The converter did not find any conversion data files.  Please check to see if the one you want is available for download at <a href="http://www.simplemachines.org/">www.simplemachines.org</a>.  If it isn\'t, we may be able to write one for you - just ask us!</h3>

			After you download it, simply upload it into the same folder as <b>this convert.php file</b>.  If you\'re having any other problems with this converter, don\'t hesitate to look for help on our <a href="http://www.simplemachines.org/community/index.php">forum</a>.<br />
			<br />
			<a href="', $_SERVER['PHP_SELF'], '?convert_script=">Try again</a>';
	}

	echo '
		</div>';

	return true;
}

function doStep0($error_message = null)
{
	global $convert_data;

	if (find_convert_scripts())
		return true;

	// If these aren't set (from an error..) default to the current directory.
	if (!isset($_POST['path_from']))
		$_POST['path_from'] = dirname(__FILE__);
	if (!isset($_POST['path_to']))
		$_POST['path_to'] = dirname(__FILE__);

	$test_from = empty($convert_data['settings']);
	foreach ($convert_data['settings'] as $s)
		$test_from |= file_exists($_POST['path_from'] . $s);

	$test_to = file_exists($_POST['path_to'] . '/Settings.php');

	// Do we have steps?
	$steps = findSteps();
	// How about charater sets.
	$charsets = array(
		// Chinese-traditional.
		'big5' => 'big5',
		// Chinese-simplified.
		'gbk' => 'gbk',
		// West European.
		'ISO-8859-1' => 'latin1',
		// Romanian.
		'ISO-8859-2' => 'latin2',
		// Turkish.
		'ISO-8859-9' => 'latin5',
		// West European with Euro sign.
		'ISO-8859-15' => 'latin9',
		// Thai.
		'tis-620' => 'tis620',
		// Persian, Chinese, etc.
		'UTF-8' => 'utf8',
		// Russian.
		'windows-1251' => 'cp1251',
		// Greek.
		'windows-1253' => 'utf8',
		// Hebrew.
		'windows-1255' => 'utf8',
		// Arabic.
		'windows-1256' => 'cp1256',
	);

	// Was an error message specified?
	if ($error_message !== null)
		echo '
			<div class="error_message">
				<div style="color: red;">', $error_message, '</div>
			</div>
			<br />';

	echo '
			<div class="panel">
				<form action="', $_SERVER['PHP_SELF'], '?step=1', isset($_REQUEST['debug']) ? '&amp;debug=' . $_REQUEST['debug'] : '', '" method="post">
					<h2>Before you continue</h2>
					<div style="margin-bottom: 2ex;">This converter assumes you have already installed SMF and that your installation is working properly.  It copies posts and data from your &quot;source&quot; installation of ', $convert_data['name'], ' into SMF, so it won\'t work without an installation of SMF.  All or some of the data in your installation of SMF will be <b>overwritten</b>.</div>';

	if (empty($convert_data['flatfile']))
		echo '
					<div style="margin-bottom: 2ex;">If the two softwares are installed in separate directories, the MySQL account SMF was installed using will need access to the other database.  Either way, both must be installed on the same MySQL server.</div>';

	echo '

					<h2>Where are they?</h2>
					<h3>The converter should only need to know where the two installations are, after which it should be able to handle everything for itself.</h3>

					<table width="100%" cellpadding="0" cellspacing="0" border="0" align="center">
						<tr>
							<td width="20%" valign="top" class="textbox"><label for="path_to">Path to SMF:</label></td>
							<td style="padding-bottom: 1ex;">
								<input type="text" name="path_to" id="path_to" value="', $_POST['path_to'], '" size="60" />
								<div style="font-style: italic; font-size: smaller;">', $test_to ? 'This may be the right path.' : 'You will need to change the value in this box.', '</div>
							</td>';

	if (!empty($convert_data['settings']))
		echo '
						</tr><tr>
							<td valign="top" class="textbox"><label for="path_from">Path to ', $convert_data['name'], ':</label></td>
							<td style="padding-bottom: 1ex;">
								<input type="text" name="path_from" id="path_from" value="', $_POST['path_from'], '" size="60" />
								<div style="font-style: italic; font-size: smaller;">', $test_from ? 'This may be the right path.' : 'You will need to change the value in this box.', '</div>
							</td>';

	if (!empty($convert_data['parameters']))
	{
		foreach ($convert_data['parameters'] as $param)
		{
			echo '
						</tr><tr>';

			if ($param['type'] == 'text')
				echo '
							<td valign="top" class="textbox"><label for="', $param['id'], '">', $param['label'], ':</label></td>
							<td style="padding-bottom: 1ex;">
								<input type="text" name="', $param['id'], '" id="', $param['id'], '" value="" size="60" />';
			elseif ($param['type'] == 'checked' || $param['type'] == 'checkbox')
				echo '
							<td valign="top" class="textbox"></td>
							<td style="padding-bottom: 1ex;">
								<input type="hidden" name="', $param['id'], '" value="0" />
								<label for="', $param['id'], '"><input type="checkbox" name="', $param['id'], '" id="', $param['id'], '" value="1"', $param['type'] == 'checked' ? ' checked="checked"' : '', ' /> ', $param['label'], '</label>';

			echo '
							</td>';
		}
	}

	echo '
						</tr><tr>
							<td valign="top" class="textbox" style="padding-top: 2ex;"><label for="db_pass">SMF database password:</label></td>
							<td valign="top" style="padding-top: 2ex; padding-bottom: 1ex;">
								<input type="password" name="db_pass" size="30" class="text" />
								<div style="font-style: italic; font-size: smaller;">The MySQL password (for verification only.)</div>
							</td>
						</tr>';

	// Now with this UTF-8 stuff we need to work more on charsets.
	if (!empty($charsets))
	{
		echo '
						<tr>
							<td valign="top" class="textbox"><label for="charsets">Set SMF\'s default character set to:</label></td>
							<td valign="top" style="padding-top: 2ex; padding-bottom: 1ex;">
								<select name="charsets" id="charsets">';
		foreach ($charsets as $name => $id)
			echo '
									<option value="', $id, '"', $id == 'latin1' ? ' selected="selected"' : '', '>', $name, '</option>';
		echo '
								</select>
							</td>
						</tr>';
	}

	// Now for the steps.
	// !!! For now this should be disabled.
	if (!empty($steps) && 1==0)
	{
		echo '
						<tr>
							<td valign="top" class="textbox">Convert selected areas ONLY:</td>
							<td valign="top" style="padding-top: 2ex; padding-bottom: 1ex;">';
		foreach ($steps as $key => $step)
			echo '
								<input type="checkbox" name="do_steps[', $step['next_step_line'], ']" id="do_steps[', $step['next_step_line'], ']" value="', $step['count'], '" checked="checked" /><label for="do_steps[', $step['next_step_line'], ']">', ucfirst(str_replace('Converting ', '', $step['name'])), '</label><br />';
		echo '
							</td>
						</tr>';
	}

	// Empty our error log?
	echo '
						</tr><tr>
							<td valign="top" class="textbox" style="padding-top: 2ex;"><label for="empty_error_log">Empty the convert error log?</label></td>
							<td valign="top" style="padding-top: 2ex; padding-bottom: 1ex;">
								<input type="checkbox" name="empty_error_log" value="1"/>
							</td>
						</tr>';

	echo '
					</table>
					<div align="right" style="margin: 1ex; margin-top: 0;"><input name="b" type="submit" value="Continue" class="submit" /></div>
				</form>
			</div>';

	if (!empty($GLOBALS['possible_scripts']))
		echo '
			<div class="panel">
				<h2>Not this software?</h2>
				<h3>If this is the wrong software, you can go back and <a href="', $_SERVER['PHP_SELF'], '?convert_script=">pick a different data file</a>.</h3>
			</div>';

	if ($error_message !== null)
	{
		show_footer();
		exit;
	}

	return;
}

// Looks at the converter and returns the steps that it's able to make.
function findSteps()
{
	global $current_type;

	// For now just SQL file support.
	if (substr($_SESSION['convert_script'], -4) == '.php' || empty($_SESSION['convert_script']))
		return;

	// Load the file.
	$lines = file(dirname(__FILE__) . '/' . $_SESSION['convert_script']);

	// Need an outside counter for the steps.
	$count_steps = 1;
	$steps = array();
	foreach ($lines as $line_number => $line)
	{
		// Get rid of any comments in the beginning of the line...
		if (substr(trim($line), 0, 2) === '/*')
			$line = preg_replace('~/\*.+?\*/~', '', $line);

		if (trim($line) === '')
			continue;

		// We found the treasure :P.
		if (substr($line, 0, 4) === '--- ')
		{
			$steps[$count_steps] = array(
				'name' => trim(substr(htmlspecialchars($line), 4)),
				'cur_step_line' => $line_number,
				'prev_step_line' => 0,
				'next_step_line' => 0,
				'count' => $count_steps,
			);

			// Previous step line number.
			if (isset($steps[$count_steps - 1]))
			{
				$steps[$count_steps]['prev_step_line'] = $steps[$count_steps - 1]['cur_step_line'];
				$steps[$count_steps - 1]['next_step_line'] = $line_number;
			}
			$count_steps++;
		}
		else
			continue;
	}

	return $steps;
}

function findSupportedCharsets()
{
	// The character sets used in SMF's language files with their db equivalent.
	$charsets = array(
		// Chinese-traditional.
		'big5' => 'big5',
		// Chinese-simplified.
		'gbk' => 'gbk',
		// West European.
		'ISO-8859-1' => 'latin1',
		// Romanian.
		'ISO-8859-2' => 'latin2',
		// Turkish.
		'ISO-8859-9' => 'latin5',
		// West European with Euro sign.
		'ISO-8859-15' => 'latin9',
		// Thai.
		'tis-620' => 'tis620',
		// Persian, Chinese, etc.
		'UTF-8' => 'utf8',
		// Russian.
		'windows-1251' => 'cp1251',
		// Greek.
		'windows-1253' => 'utf8',
		// Hebrew.
		'windows-1255' => 'utf8',
		// Arabic.
		'windows-1256' => 'cp1256',
	);

	// Get a list of character sets supported by your MySQL server.
	$result = @mysql_query("
		SHOW CHARACTER SET;");
	$db_charsets = array();
	while ($row = mysql_fetch_assoc($result))
		$db_charsets[] = $row['Charset'];

	// Character sets supported by both MySQL and SMF's language files.
	return $charsets = array_intersect($charsets, $db_charsets);
}

function doStep1()
{
	global $from_prefix, $to_prefix, $convert_data, $command_line;

	if (substr($_SESSION['convert_script'], -4) == '.php')
		return run_php_converter();

	foreach ($convert_data['globals'] as $global)
		global $$global;

	$_GET['substep'] = (int) @$_GET['substep'];

	// In CLI we end lines differently.
	$endl = $command_line ? "\n" : '<br />' . "\n";

	// Staring the converter.
	if ($command_line)
		print_line($endl . 'Starting Conversion' . $endl);

	$lines = file(dirname(__FILE__) . '/' . $_SESSION['convert_script']);

	// Available steps.  This is the list of all the steps that were found.
	$available_steps = findSteps();

	$current_type = 'sql';
	$current_data = '';
	$substep = 0;
	$last_step = '';
	$special_table = null;
	$special_code = null;

	foreach ($lines as $line_number => $line)
	{
		$do_current = $substep >= $_GET['substep'];

		// Get rid of any comments in the beginning of the line...
		if (substr(trim($line), 0, 2) === '/*')
			$line = preg_replace('~/\*.+?\*/~', '', $line);

		if (trim($line) === '')
			continue;

		// Skipping steps?
		/*if (isset($_SESSION['do_steps']))
		{
			$current_step1 = trim(substr($line, 4));
			$do_steps = $_SESSION['do_steps'];
			$reverse = array_flip($_SESSION['do_steps']);
			//print_r($_SESSION['do_steps']);die;
			//if (isset($reverse[$current_step1]) && (!in_array($current_step1, $do_steps) && ($line_number < $reverse[$current_step1])))
			//	continue;
			//echo $substep;
			//print_r($do_steps);
			if (!in_array($substep, $do_steps))
			{
				// Lets make sure $substep is not 0
				$substep = $substep == 0 ? 1 : $subtep++;
				print_line($current_step1 . ' skipped...' . $endl);
				$_REQUEST['start'] = 0;
				pastTime($substep);
			}
		}*/

		if (trim(substr($line, 0, 3)) === '---')
		{
			$type = substr($line, 3, 1);

			if (trim($current_data) != '' && $type !== '}')
				print_line('Error in convert script - line ' . $line_number . '!' . $endl);

			if ($type == ' ')
			{
				if ($do_current && $_GET['substep'] != 0)
				{
					print_line(' Successful.' . $endl);
					flush();
				}

				$last_step = htmlspecialchars(rtrim(substr($line, 4)));

				if ($do_current)
				{
					$print_line = ($command_line ? ' * ' : '') . $last_step . (empty($_SESSION['convert_debug']) ? '' : $endl);
					print_line($print_line);

					pastTime($substep);
				}
			}
			elseif ($type == '#')
			{
				if (!empty($_SESSION['convert_debug']) && $do_current)
				{
					if (trim($line) == '---#')
						print_line(' done.' . $endl);
					else
						echo '&nbsp;&nbsp;&nbsp;', htmlspecialchars(rtrim(substr($line, 4)));
				}

				if ($substep < $_GET['substep'] && $substep + 1 >= $_GET['substep'])
				{
					$print_line = ($command_line ? ' * ' : '') . $last_step . (empty($_SESSION['convert_debug']) ? '' : $endl);
					print_line($print_line);
				}

				// Small step!
				pastTime(++$substep);
			}
			elseif ($type == '{')
				$current_type = 'code';
			elseif ($type == '}')
			{
				$current_type = 'sql';

				if (!$do_current)
				{
					$current_data = '';
					continue;
				}

				if ($special_table !== null)
					$special_code = $current_data;
				else
				{
					if (eval($current_data) === false)
					{
						if (!$command_line)
							echo '
			<b>Error in convert script ', $_SESSION['convert_script'], ' on line ', $line_number, '!</b><br />';
						else
							print_line('Error in convert script ' . $_SESSION['convert_script'] . ' on line ' . $line_number . $endl);
					}
				}

				// Done with code!
				$current_data = '';
			}
			elseif ($type == '*')
			{
				if ($substep < $_GET['substep'] && $substep + 1 >= $_GET['substep'])
				{
					$print_line = $last_step . (empty($_SESSION['convert_debug']) ? ' ' : $endl);
					print_line($print_line);
				}

				if ($special_table === null)
				{
					$special_table = strtr(trim(substr($line, 4)), array('{$to_prefix}' => $to_prefix));

					if (preg_match('~^([^ ()]+?)( \(update .+?\))? (\d+)$~', trim($special_table), $match) != 0)
					{
						$special_table = $match[1];
						$special_update = $match[2] != '' ? substr($match[2], 9, -1) : '';
						$special_limit = empty($match[3]) ? 500 : (int) $match[3];
					}
					elseif (preg_match('~^([^ ()]+?) \(update (.+?)\)$~', trim($special_table), $match) != 0)
					{
						$special_table = $match[1];
						$special_update = $match[2];
						$special_limit = 200;
					}
					else
					{
						$special_update = false;
						$special_limit = 500;
					}
				}
				else
				{
					$special_table = null;
					$special_code = null;
				}

				// Increase the substep slightly...
				pastTime(++$substep);
			}

			continue;
		}

		$current_data .= $line;
		if (substr(rtrim($current_data), -1) === ';' && $current_type === 'sql')
		{
			if (!$do_current)
			{
				$current_data = '';
				continue;
			}

			$current_data = strtr(substr(rtrim($current_data), 0, -1), array('{$from_prefix}' => $from_prefix, '{$to_prefix}' => $to_prefix));
			if (strpos($current_data, '{$') !== false)
				$current_data = eval('return "' . addcslashes($current_data, '\\"') . '";');

			if (isset($convert_table) && $convert_table !== null && strpos($current_data, '%d') !== false)
			{
				preg_match('~FROM [(]?([^\s,]+)~i', $convert_data, $match);
				if (!empty($match))
				{
					$result = convert_query("
						SELECT COUNT(*)
						FROM $match[1]");
					list ($special_max) = convert_fetch_row($result);
					mysql_free_result($result);
				}
				else
					$special_max = 0;
			}
			else
				$special_max = 0;

			if ($special_table === null)
				convert_query($current_data);
			elseif ($special_update != false)
			{
				while (true)
				{
					pastTime($substep);

					if (strpos($current_data, '%d') !== false)
						$special_result = convert_query(sprintf($current_data, $_REQUEST['start'], $_REQUEST['start'] + $special_limit - 1) . "\n" . 'LIMIT ' . $special_limit);
					else
						$special_result = convert_query($current_data . "\n" . 'LIMIT ' . $_REQUEST['start'] . ', ' . $special_limit);
					while ($row = convert_fetch_assoc($special_result))
					{
						if ($special_code !== null)
							eval($special_code);

						if (empty($no_add) && count($row) >= 2)
						{
							$setString = array();
							foreach ($row as $k => $v)
							{
								if ($k != $special_update)
									$setString[] = "$k = '" . addslashes($v) . "'";
							}

							convert_query("
								UPDATE " . $special_table . "
								SET " . implode(', ', $setString) . "
								WHERE $special_update = '" . $row[$special_update] . "'");
						}
						else
							$no_add = false;
					}

					$_REQUEST['start'] += $special_limit;
					if (empty($special_max) && convert_num_rows($special_result) < $special_limit)
						break;
					elseif (!empty($special_max) && convert_num_rows($special_result) == 0 && $_REQUEST['start'] > $special_max)
						break;
					convert_free_result($special_result);
				}
			}
			else
			{
				// Are we doing attachments?  They're going to want a few things...
				if ($special_table == $to_prefix . 'attachments')
				{
					if (!isset($ID_ATTACH, $attachmentUploadDir))
					{
						$result = convert_query("
							SELECT MAX(ID_ATTACH) + 1
							FROM {$to_prefix}attachments");
						list ($ID_ATTACH) = mysql_fetch_row($result);
						mysql_free_result($result);

						$result = convert_query("
							SELECT value
							FROM {$to_prefix}settings
							WHERE variable = 'attachmentUploadDir'
							LIMIT 1");
						list ($attachmentUploadDir) = mysql_fetch_row($result);
						mysql_free_result($result);

						if (empty($ID_ATTACH))
							$ID_ATTACH = 1;
					}
				}

				while (true)
				{
					pastTime($substep);

					if (strpos($current_data, '%d') !== false)
						$special_result = convert_query(sprintf($current_data, $_REQUEST['start'], $_REQUEST['start'] + $special_limit - 1) . "\n" . 'LIMIT ' . $special_limit);
					else
						$special_result = convert_query($current_data . "\n" . 'LIMIT ' . $_REQUEST['start'] . ', ' . $special_limit);
					$rows = array();
					$keys = array();
					while ($row = convert_fetch_assoc($special_result))
					{
						if ($special_code !== null)
							eval($special_code);

						// Here we have various bits of custom code for some known problems global to all converters.
						if ($special_table == $to_prefix . 'members')
						{
							// Let's ensure there are no illegal characters.
							$row['memberName'] = preg_replace('/[<>&"\'=\\\]/is', '', $row['memberName']);
							$row['realName'] = trim($row['realName'], " \t\n\r\x0B\0\xA0");

							if (strpos($row['realName'], '<') !== false || strpos($row['realName'], '>') !== false || strpos($row['realName'], '& ') !== false)
								$row['realName'] = htmlspecialchars($row['realName'], ENT_QUOTES);
							else
								$row['realName'] = strtr($row['realName'], array('\'' => '&#039;'));
						}

						if (empty($no_add))
							$rows[] = "'" . implode("', '", addslashes_recursive($row)) . "'";
						else
							$no_add = false;

						if (empty($keys))
							$keys = array_keys($row);
					}

					if (isset($ignore) && $ignore == true)
						$ignore = 'IGNORE';
					else
						$ignore = '';

					if (!empty($rows))
						convert_query("
							INSERT $ignore INTO $special_table
								(" . implode(', ', $keys) . ")
							VALUES (" . implode('),
								(', $rows) . ")");

					$_REQUEST['start'] += $special_limit;
					if (empty($special_max) && convert_num_rows($special_result) < $special_limit)
						break;
					elseif (!empty($special_max) && convert_num_rows($special_result) == 0 && $_REQUEST['start'] > $special_max)
						break;
					convert_free_result($special_result);
				}
			}

			$_REQUEST['start'] = 0;
			$special_code = null;
			$current_data = '';
		}
	}

	print_line(' Successful.' . $endl);
	flush();

	$_GET['substep'] = 0;
	$_REQUEST['start'] = 0;

	return doStep2();
}

function run_php_converter()
{
	global $from_prefix, $to_prefix, $convert_data, $command_line;

	foreach ($convert_data['globals'] as $global)
		global $$global;

	$_GET['substep'] = (int) @$_GET['substep'];
	$_GET['cstep'] = (int) @$_GET['cstep'];

	require(dirname(__FILE__) . '/' . $_SESSION['convert_script']);

	// CLI uses \n.
	$endl = $command_line ? "\n" : '<br />' . "\n";

	if (function_exists('load_converter_settings'))
		load_converter_settings();

	for ($_GET['cstep'] = max(1, $_GET['cstep']); function_exists('convertStep' . $_GET['cstep']); $_GET['cstep']++)
	{
		call_user_func('convertStep' . $_GET['cstep']);
		$_GET['substep'] = 0;
		pastTime(0);

		print_line(' Successful.' . $endl);
		flush();
	}

	$_GET['substep'] = 0;
	$_REQUEST['start'] = 0;

	return doStep2();
}

function doStep2()
{
	global $convert_data, $from_prefix, $to_prefix, $modSettings, $command_line;

	$_GET['step'] = '2';

	// CLI uses \n.
	$endl = $command_line ? "\n" : '<br />' . "\n";

	print_line(($command_line ? ' * ' : '') . 'Recalculating forum statistics... ');

	if ($_GET['substep'] <= 0)
	{
		// Get all members with wrong number of personal messages.
		$request = convert_query("
			SELECT mem.ID_MEMBER, COUNT(pmr.ID_PM) AS realNum, mem.instantMessages
			FROM {$to_prefix}members AS mem
				LEFT JOIN {$to_prefix}pm_recipients AS pmr ON (mem.ID_MEMBER = pmr.ID_MEMBER AND pmr.deleted = 0)
			GROUP BY mem.ID_MEMBER
			HAVING realNum != instantMessages");
		while ($row = mysql_fetch_assoc($request))
		{
			convert_query("
				UPDATE {$to_prefix}members
				SET instantMessages = $row[realNum]
				WHERE ID_MEMBER = $row[ID_MEMBER]
				LIMIT 1");

			pastTime(0);
		}
		mysql_free_result($request);

		$request = convert_query("
			SELECT mem.ID_MEMBER, COUNT(pmr.ID_PM) AS realNum, mem.unreadMessages
			FROM {$to_prefix}members AS mem
				LEFT JOIN {$to_prefix}pm_recipients AS pmr ON (mem.ID_MEMBER = pmr.ID_MEMBER AND pmr.deleted = 0 AND pmr.is_read = 0)
			GROUP BY mem.ID_MEMBER
			HAVING realNum != unreadMessages");
		while ($row = mysql_fetch_assoc($request))
		{
			convert_query("
				UPDATE {$to_prefix}members
				SET unreadMessages = $row[realNum]
				WHERE ID_MEMBER = $row[ID_MEMBER]
				LIMIT 1");

			pastTime(0);
		}
		mysql_free_result($request);

		pastTime(1);
	}

	if ($_GET['substep'] <= 1)
	{
		$request = convert_query("
			SELECT ID_BOARD, MAX(ID_MSG) AS ID_LAST_MSG, MAX(modifiedTime) AS lastEdited
			FROM {$to_prefix}messages
			GROUP BY ID_BOARD");
		$modifyData = array();
		$modifyMsg = array();
		while ($row = mysql_fetch_assoc($request))
		{
			convert_query("
				UPDATE {$to_prefix}boards
				SET ID_LAST_MSG = $row[ID_LAST_MSG], ID_MSG_UPDATED = $row[ID_LAST_MSG]
				WHERE ID_BOARD = $row[ID_BOARD]
				LIMIT 1");
			$modifyData[$row['ID_BOARD']] = array(
				'last_msg' => $row['ID_LAST_MSG'],
				'lastEdited' => $row['lastEdited'],
			);
			$modifyMsg[] = $row['ID_LAST_MSG'];
		}
		mysql_free_result($request);

		// Are there any boards where the updated message is not the last?
		if (!empty($modifyMsg))
		{
			$request = convert_query("
				SELECT ID_BOARD, ID_MSG, modifiedTime, posterTime
				FROM {$to_prefix}messages
				WHERE ID_MSG IN (" . implode(',', $modifyMsg) . ")");
			while ($row = mysql_fetch_assoc($request))
			{
				// Have we got a message modified before this was posted?
				if (max($row['modifiedTime'], $row['posterTime']) < $modifyData[$row['ID_BOARD']]['lastEdited'])
				{
					// Work out the ID of the message (This seems long but it won't happen much.
					$request2 = convert_query("
						SELECT ID_MSG
						FROM {$to_prefix}messages
						WHERE modifiedTime = " . $modifyData[$row['ID_BOARD']]['lastEdited'] . "
						LIMIT 1");
					if (mysql_num_rows($request2) != 0)
					{
						list ($ID_MSG) = mysql_fetch_row($request2);

						convert_query("
							UPDATE {$to_prefix}boards
							SET ID_MSG_UPDATED = $ID_MSG
							WHERE ID_BOARD = $row[ID_BOARD]
							LIMIT 1");
					}
					mysql_free_result($request2);
				}
			}
			mysql_free_result($request);
		}

		pastTime(2);
	}

	if ($_GET['substep'] <= 2)
	{
		$request = convert_query("
			SELECT ID_GROUP
			FROM {$to_prefix}membergroups
			WHERE minPosts = -1");
		$all_groups = array();
		while ($row = mysql_fetch_assoc($request))
			$all_groups[] = $row['ID_GROUP'];
		mysql_free_result($request);

		$request = convert_query("
			SELECT ID_BOARD, memberGroups
			FROM {$to_prefix}boards
			WHERE FIND_IN_SET(0, memberGroups)");
		while ($row = mysql_fetch_assoc($request))
			convert_query("
				UPDATE {$to_prefix}boards
				SET memberGroups = '" . implode(',', array_unique(array_merge($all_groups, explode(',', $row['memberGroups'])))) . "'
				WHERE ID_BOARD = $row[ID_BOARD]
				LIMIT 1");
		mysql_free_result($request);

		pastTime(3);
	}

	if ($_GET['substep'] <= 3)
	{
		// Get the number of messages...
		$result = convert_query("
			SELECT COUNT(*) AS totalMessages, MAX(ID_MSG) AS maxMsgID
			FROM {$to_prefix}messages");
		$row = mysql_fetch_assoc($result);
		mysql_free_result($result);

		// Update the latest member.  (highest ID_MEMBER)
		$result = convert_query("
			SELECT ID_MEMBER AS latestMember, realName AS latestRealName
			FROM {$to_prefix}members
			ORDER BY ID_MEMBER DESC
			LIMIT 1");
		if (mysql_num_rows($result))
			$row += mysql_fetch_assoc($result);
		mysql_free_result($result);

		// Update the member count.
		$result = convert_query("
			SELECT COUNT(*) AS totalMembers
			FROM {$to_prefix}members");
		$row += mysql_fetch_assoc($result);
		mysql_free_result($result);

		// Get the number of topics.
		$result = convert_query("
			SELECT COUNT(*) AS totalTopics
			FROM {$to_prefix}topics");
		$row += mysql_fetch_assoc($result);
		mysql_free_result($result);

		convert_query("
			REPLACE INTO {$to_prefix}settings
				(variable, value)
			VALUES ('latestMember', '$row[latestMember]'),
				('latestRealName', '$row[latestRealName]'),
				('totalMembers', '$row[totalMembers]'),
				('totalMessages', '$row[totalMessages]'),
				('maxMsgID', '$row[maxMsgID]'),
				('totalTopics', '$row[totalTopics]'),
				('disableHashTime', " . (time() + 7776000) . ")");

		pastTime(4);
	}

	if ($_GET['substep'] <= 4)
	{
		$request = convert_query("
			SELECT ID_GROUP, minPosts
			FROM {$to_prefix}membergroups
			WHERE minPosts != -1
			ORDER BY minPosts DESC");
		$post_groups = array();
		while ($row = mysql_fetch_assoc($request))
			$post_groups[$row['minPosts']] = $row['ID_GROUP'];
		mysql_free_result($request);

		$request = convert_query("
			SELECT ID_MEMBER, posts
			FROM {$to_prefix}members");
		$mg_updates = array();
		while ($row = mysql_fetch_assoc($request))
		{
			$group = 4;
			foreach ($post_groups as $min_posts => $group_id)
				if ($row['posts'] >= $min_posts)
				{
					$group = $group_id;
					break;
				}

			$mg_updates[$group][] = $row['ID_MEMBER'];
		}
		mysql_free_result($request);

		foreach ($mg_updates as $group_to => $update_members)
			convert_query("
				UPDATE {$to_prefix}members
				SET ID_POST_GROUP = $group_to
				WHERE ID_MEMBER IN (" . implode(', ', $update_members) . ")
				LIMIT " . count($update_members));

		// This isn't completely related, but should be rather quick.
		convert_query("
			UPDATE {$to_prefix}members
			SET ICQ = ''
			WHERE ICQ = '0'");

		pastTime(5);
	}

	if ($_GET['substep'] <= 5)
	{
		// Needs to be done separately for each board.
		$result_boards = convert_query("
			SELECT ID_BOARD
			FROM {$to_prefix}boards");
		$boards = array();
		while ($row_boards = mysql_fetch_assoc($result_boards))
			$boards[] = $row_boards['ID_BOARD'];
		mysql_free_result($result_boards);

		foreach ($boards as $ID_BOARD)
		{
			// Get the number of topics, and iterate through them.
			$result_topics = convert_query("
				SELECT COUNT(*)
				FROM {$to_prefix}topics
				WHERE ID_BOARD = $ID_BOARD");
			list ($num_topics) = mysql_fetch_row($result_topics);
			mysql_free_result($result_topics);

			// Find how many messages are in the board.
			$result_posts = convert_query("
				SELECT COUNT(*)
				FROM {$to_prefix}messages
				WHERE ID_BOARD = $ID_BOARD");
			list ($num_posts) = mysql_fetch_row($result_posts);
			mysql_free_result($result_posts);

			// Fix the board's totals.
			convert_query("
				UPDATE {$to_prefix}boards
				SET numTopics = $num_topics, numPosts = $num_posts
				WHERE ID_BOARD = $ID_BOARD
				LIMIT 1");
		}

		pastTime(6);
	}

	// Remove all topics that have zero messages in the messages table.
	if ($_GET['substep'] <= 6)
	{
		while (true)
		{
			$resultTopic = convert_query("
				SELECT t.ID_TOPIC, COUNT(m.ID_MSG) AS numMsg
				FROM {$to_prefix}topics AS t
					LEFT JOIN {$to_prefix}messages AS m ON (m.ID_TOPIC = t.ID_TOPIC)
				GROUP BY t.ID_TOPIC
				HAVING numMsg = 0
				LIMIT $_REQUEST[start], 200");

			$numRows = mysql_num_rows($resultTopic);

			if ($numRows > 0)
			{
				$stupidTopics = array();
				while ($topicArray = mysql_fetch_assoc($resultTopic))
					$stupidTopics[] = $topicArray['ID_TOPIC'];
				convert_query("
					DELETE FROM {$to_prefix}topics
					WHERE ID_TOPIC IN (" . implode(',', $stupidTopics) . ')
					LIMIT ' . count($stupidTopics));
				convert_query("
					DELETE FROM {$to_prefix}log_topics
					WHERE ID_TOPIC IN (" . implode(',', $stupidTopics) . ')');
			}
			mysql_free_result($resultTopic);

			if ($numRows < 200)
				break;

			$_REQUEST['start'] += 200;
			pastTime(6);
		}

		$_REQUEST['start'] = 0;
		pastTime(7);
	}

	// Get the correct number of replies.
	if ($_GET['substep'] <= 7)
	{
		while (true)
		{
			$resultTopic = convert_query("
				SELECT
					t.ID_TOPIC, MIN(m.ID_MSG) AS myID_FIRST_MSG, t.ID_FIRST_MSG,
					MAX(m.ID_MSG) AS myID_LAST_MSG, t.ID_LAST_MSG, COUNT(m.ID_MSG) - 1 AS myNumReplies,
					t.numReplies
				FROM {$to_prefix}topics AS t
					LEFT JOIN {$to_prefix}messages AS m ON (m.ID_TOPIC = t.ID_TOPIC)
				GROUP BY t.ID_TOPIC
				HAVING ID_FIRST_MSG != myID_FIRST_MSG OR ID_LAST_MSG != myID_LAST_MSG OR numReplies != myNumReplies
				LIMIT $_REQUEST[start], 200");

			$numRows = mysql_num_rows($resultTopic);

			while ($topicArray = mysql_fetch_assoc($resultTopic))
			{
				$memberStartedID = getMsgMemberID($topicArray['myID_FIRST_MSG']);
				$memberUpdatedID = getMsgMemberID($topicArray['myID_LAST_MSG']);

				convert_query("
					UPDATE {$to_prefix}topics
					SET ID_FIRST_MSG = '$topicArray[myID_FIRST_MSG]',
						ID_MEMBER_STARTED = '$memberStartedID', ID_LAST_MSG = '$topicArray[myID_LAST_MSG]',
						ID_MEMBER_UPDATED = '$memberUpdatedID', numReplies = '$topicArray[myNumReplies]'
					WHERE ID_TOPIC = $topicArray[ID_TOPIC]
					LIMIT 1");
			}
			mysql_free_result($resultTopic);

			if ($numRows < 200)
				break;

			$_REQUEST['start'] += 100;
			pastTime(7);
		}

		$_REQUEST['start'] = 0;
		pastTime(8);
	}

	// Fix ID_CAT, ID_PARENT, and childLevel.
	if ($_GET['substep'] <= 8)
	{
		// First, let's get an array of boards and parents.
		$request = convert_query("
			SELECT ID_BOARD, ID_PARENT, ID_CAT
			FROM {$to_prefix}boards");
		$child_map = array();
		$cat_map = array();
		while ($row = mysql_fetch_assoc($request))
		{
			$child_map[$row['ID_PARENT']][] = $row['ID_BOARD'];
			$cat_map[$row['ID_BOARD']] = $row['ID_CAT'];
		}
		mysql_free_result($request);

		// Let's look for any boards with obviously invalid parents...
		foreach ($child_map as $parent => $dummy)
		{
			if ($parent != 0 && !isset($cat_map[$parent]))
			{
				// Perhaps it was supposed to be their ID_CAT?
				foreach ($dummy as $board)
				{
					if (empty($cat_map[$board]))
						$cat_map[$board] = $parent;
				}

				$child_map[0] = array_merge(isset($child_map[0]) ? $child_map[0] : array(), $dummy);
				unset($child_map[$parent]);
			}
		}

		// The above ID_PARENTs and ID_CATs may all be wrong; we know ID_PARENT = 0 is right.
		$solid_parents = array(array(0, 0));
		$fixed_boards = array();
		while (!empty($solid_parents))
		{
			list ($parent, $level) = array_pop($solid_parents);
			if (!isset($child_map[$parent]))
				continue;

			// Fix all of this board's children.
			foreach ($child_map[$parent] as $board)
			{
				if ($parent != 0)
					$cat_map[$board] = $cat_map[$parent];
				$fixed_boards[$board] = array($parent, $cat_map[$board], $level);
				$solid_parents[] = array($board, $level + 1);
			}
		}

		foreach ($fixed_boards as $board => $fix)
		{
			convert_query("
				UPDATE {$to_prefix}boards
				SET ID_PARENT = " . (int) $fix[0] . ", ID_CAT = " . (int) $fix[1] . ", childLevel = " . (int) $fix[2] . "
				WHERE ID_BOARD = " . (int) $board . "
				LIMIT 1");
		}

		// Leftovers should be brought to the root.  They had weird parents we couldn't find.
		if (count($fixed_boards) < count($cat_map))
		{
			convert_query("
				UPDATE {$to_prefix}boards
				SET childLevel = 0, ID_PARENT = 0" . (empty($fixed_boards) ? '' : "
				WHERE ID_BOARD NOT IN (" . implode(', ', array_keys($fixed_boards)) . ")"));
		}

		// Last check: any boards not in a good category?
		$request = convert_query("
			SELECT ID_CAT
			FROM {$to_prefix}categories");
		$real_cats = array();
		while ($row = mysql_fetch_assoc($request))
			$real_cats[] = $row['ID_CAT'];
		mysql_free_result($request);

		$fix_cats = array();
		foreach ($cat_map as $board => $cat)
		{
			if (!in_array($cat, $real_cats))
				$fix_cats[] = $cat;
		}

		if (!empty($fix_cats))
		{
			convert_query("
				INSERT INTO {$to_prefix}categories
					(name)
				VALUES ('General Category')");
			$catch_cat = mysql_insert_id();

			convert_query("
				UPDATE {$to_prefix}boards
				SET ID_CAT = " . (int) $catch_cat . "
				WHERE ID_CAT IN (" . implode(', ', array_unique($fix_cats)) . ")");
		}

		pastTime(9);
	}

	if ($_GET['substep'] <= 9)
	{
		$request = convert_query("
			SELECT c.ID_CAT, c.catOrder, b.ID_BOARD, b.boardOrder
			FROM {$to_prefix}categories AS c
				LEFT JOIN {$to_prefix}boards AS b ON (b.ID_CAT = c.ID_CAT)
			ORDER BY c.catOrder, b.childLevel, b.boardOrder, b.ID_BOARD");
		$catOrder = -1;
		$boardOrder = -1;
		$curCat = -1;
		while ($row = mysql_fetch_assoc($request))
		{
			if ($curCat != $row['ID_CAT'])
			{
				$curCat = $row['ID_CAT'];
				if (++$catOrder != $row['catOrder'])
					convert_query("
						UPDATE {$to_prefix}categories
						SET catOrder = $catOrder
						WHERE ID_CAT = $row[ID_CAT]
						LIMIT 1");
			}
			if (!empty($row['ID_BOARD']) && ++$boardOrder != $row['boardOrder'])
				convert_query("
					UPDATE {$to_prefix}boards
					SET boardOrder = $boardOrder
					WHERE ID_BOARD = $row[ID_BOARD]
					LIMIT 1");
		}
		mysql_free_result($request);

		pastTime(10);
	}

	if ($_GET['substep'] <= 10)
	{
		convert_query("
			ALTER TABLE {$to_prefix}boards
			ORDER BY boardOrder");

		convert_query("
			ALTER TABLE {$to_prefix}smileys
			ORDER BY code DESC");

		pastTime(11);
	}

	if ($_GET['substep'] <= 11)
	{
		$request = convert_query("
			SELECT COUNT(*)
			FROM {$to_prefix}attachments");
		list ($attachments) = mysql_fetch_row($request);
		mysql_free_result($request);

		while ($_REQUEST['start'] < $attachments)
		{
			$request = convert_query("
				SELECT ID_ATTACH, filename, attachmentType
				FROM {$to_prefix}attachments
				WHERE ID_THUMB = 0
					AND (RIGHT(filename, 4) IN ('.gif', '.jpg', '.png', '.bmp') OR RIGHT(filename, 5) = '.jpeg')
					AND width = 0
					AND height = 0
				LIMIT $_REQUEST[start], 500");
			if (mysql_num_rows($request) == 0)
				break;
			while ($row = mysql_fetch_assoc($request))
			{
				if ($row['attachmentType'] == 1)
				{
					$request2 = convert_query("
						SELECT value
						FROM {$to_prefix}settings
						WHERE variable = 'custom_avatar_dir'
						LIMIT 1");
					list ($custom_avatar_dir) = mysql_fetch_row($request2);
					mysql_free_result($request2);

					$filename = $custom_avatar_dir . '/' . $row['filename'];
				}
				else
					$filename = getLegacyAttachmentFilename($row['filename'], $row['ID_ATTACH']);

				// Probably not one of the converted ones, then?
				if (!file_exists($filename))
					continue;

				$size = @getimagesize($filename);
				$filesize = @filesize($filename);
				if (!empty($size) && !empty($size[0]) && !empty($size[1]) && !empty($filesize))
					convert_query("
						UPDATE {$to_prefix}attachments
						SET
							size = " . (int) $filesize . ",
							width = " . (int) $size[0] . ",
							height = " . (int) $size[1] . "
						WHERE ID_ATTACH = $row[ID_ATTACH]
						LIMIT 1");
			}
			mysql_free_result($request);

			// More?
			// We can't keep converting the same files over and over again!
			$_REQUEST['start'] += 500;
			pastTime(11);
		}

		$_REQUEST['start'] = 0;
		pastTime(12);
	}

	// Lets rebuild the indexes.
	if ($_GET['substep'] <= 12)
	{
		$knownKeys = array(
			'PRIMARY' => 'ADD PRIMARY KEY (ID_TOPIC)',
			'lastMessage' => 'ADD UNIQUE lastMessage (ID_LAST_MSG, ID_BOARD)',
			'firstMessage' => 'ADD UNIQUE firstMessage (ID_FIRST_MSG, ID_BOARD)',
			'poll' => 'ADD UNIQUE poll (ID_POLL, ID_TOPIC)',
			'isSticky' => 'ADD KEY isSticky (isSticky)',
			'ID_BOARD' => 'ADD KEY ID_BOARD (ID_BOARD)',
		);
		alterTable('topics', $knownKeys, '', '', true);

		$_REQUEST['start'] = 0;
		pastTime(13);
	}

	if ($_GET['substep'] <= 13)
	{
		$knownKeys = array(
			'PRIMARY' => 'ADD PRIMARY KEY (ID_MSG)',
			'topic' => 'ADD UNIQUE topic (ID_TOPIC, ID_MSG)',
			'ID_BOARD' => 'ADD UNIQUE ID_BOARD (ID_BOARD, ID_MSG)',
			'ID_MEMBER' => 'ADD UNIQUE ID_MEMBER (ID_MEMBER, ID_MSG)',
			'ipIndex' => 'ADD KEY ipIndex (posterIP(15), ID_TOPIC)',
			'participation' => 'ADD KEY participation (ID_MEMBER, ID_TOPIC)',
			'showPosts' => 'ADD KEY showPosts (ID_MEMBER, ID_BOARD)',
			'ID_TOPIC' => 'ADD KEY ID_TOPIC (ID_TOPIC)',
		);
		alterTable('messages', $knownKeys, '', '', true);

		$_REQUEST['start'] = 0;
		pastTime(14);
	}

	print_line(' Successful.' . $endl);

	return doStep3();
}

function doStep3()
{
	global $boardurl, $convert_data, $command_line;

	// In CLI we end lines differently.
	$endl = $command_line ? "\n" : '<br />' . "\n";
	if ($command_line)
	{
		print_line($endl);
		print_line('Conversion Complete!' . $endl);
		print_line('Please delete this file as soon as possible for security reasons.' . $endl);
		return true;
	}

	echo '
				<h2 style="margin-top: 2ex;">Conversion Complete</h2>
				<h3>Congratulations, the conversion has completed successfully.  If you have or had any problems with this converter, or need help using SMF, please feel free to <a href="http://www.simplemachines.org/community/index.php">look to us for support</a>.</h3>';

	if (is_writable(dirname(__FILE__)) && is_writable(__FILE__))
		echo '
				<div style="margin: 1ex; font-weight: bold;">
					<label for="delete_self"><input type="checkbox" id="delete_self" onclick="doTheDelete();" /> Please check this box to delete the converter right now for security reasons.</label> (doesn\'t work on all servers.)
				</div>
				<script language="JavaScript" type="text/javascript"><!-- // --><![CDATA[
					function doTheDelete()
					{
						var theCheck = document.getElementById ? document.getElementById("delete_self") : document.all.delete_self;
						var tempImage = new Image();

						tempImage.src = "', $_SERVER['PHP_SELF'], '?delete=1&" + (new Date().getTime());
						tempImage.width = 0;
						theCheck.disabled = true;
					}
				// ]]></script>
				<br />';
	echo '
				Now that everything is converted over, <a href="', $boardurl, '/index.php">your SMF installation</a> should have all the posts, boards, and members from the ', $convert_data['name'], ' installation.<br />
				<br />
				We hope you had a smooth transition!';

	return true;
}

function show_header()
{
	global $convert_data, $time_start;
	$time_start = time();

	echo '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html>
	<head>
	<title>', isset($convert_data['name']) ? $convert_data['name'] . ' to ' : '', 'SMF Converter</title>
		<style type="text/css">
			body
			{
				background-color: #E5E5E8;
				margin: 0px;
				padding: 0px;
			}
			body, td
			{
				color: #000000;
				font-size: small;
				font-family: verdana, sans-serif;
			}
			div#header
			{
				background-image: url(Themes/default/images/catbg.jpg);
				background-repeat: repeat-x;
				background-color: #88A6C0;
				padding: 22px 4% 12px 4%;
				color: white;
				font-family: Georgia, serif;
				font-size: xx-large;
				border-bottom: 1px solid black;
				height: 40px;
			}
			div#content
			{
				padding: 20px 30px;
			}
			div.error_message
			{
				border: 2px dashed red;
				background-color: #E1E1E1;
				margin: 1ex 4ex;
				padding: 1.5ex;
			}
			div.panel
			{
				border: 1px solid gray;
				background-color: #F6F6F6;
				margin: 1ex 0;
				padding: 1.2ex;
			}
			div.panel h2
			{
				margin: 0;
				margin-bottom: 0.5ex;
				padding-bottom: 3px;
				border-bottom: 1px dashed black;
				font-size: 14pt;
				font-weight: normal;
			}
			div.panel h3
			{
				margin: 0;
				margin-bottom: 2ex;
				font-size: 10pt;
				font-weight: normal;
			}
			form
			{
				margin: 0;
			}
			td.textbox
			{
				padding-top: 2px;
				font-weight: bold;
				white-space: nowrap;
				padding-', empty($txt['lang_rtl']) ? 'right' : 'left', ': 2ex;
			}
		</style>
	</head>
	<body>
		<div id="header">
			<div title="Bahamut!">', isset($convert_data['name']) ? $convert_data['name'] . ' to ' : '', 'SMF Converter</div>
		</div>
		<div id="content">';
}

// Show the footer.
function show_footer()
{
	echo '
		</div>
	</body>
</html>';
}

// Check if we've passed the time limit..
function pastTime($substep = null)
{
	global $time_start, $command_line;

	if (isset($_GET['substep']) && $_GET['substep'] < $substep)
		$_GET['substep'] = $substep;

	if ($command_line)
	{
		if (time() - $time_start > 1)
		{
			print_line('.');
			$time_start = time();
		}
		return;
	}

	@set_time_limit(300);
	if (function_exists('apache_reset_timeout'))
		apache_reset_timeout();

	if (time() - $time_start < 10)
		return;

	echo '
			<i>Incomplete.</i><br />

			<h2 style="margin-top: 2ex;">Not quite done yet!</h2>
			<h3>
				This conversion has paused to avoid overloading your server, and hence not working properly.<br />
				Don\'t worry though, <b>nothing\'s wrong</b> - simply click the <label for="continue">continue button</label> below to start the converter from where it left off.
			</h3>

			<form action="', $_SERVER['PHP_SELF'], '?step=', $_GET['step'], isset($_GET['substep']) ? '&amp;substep=' . $_GET['substep'] : '', isset($_GET['cstep']) ? '&amp;cstep=' . $_GET['cstep'] : '', '&amp;start=', $_REQUEST['start'], '" method="post" name="autoSubmit">
				<div align="right" style="margin: 1ex;"><input name="b" type="submit" value="Continue" /></div>
			</form>
			<script language="JavaScript" type="text/javascript"><!-- // --><![CDATA[
				window.onload = doAutoSubmit;
				var countdown = 3;

				function doAutoSubmit()
				{
					if (countdown == 0)
						document.autoSubmit.submit();
					else if (countdown == -1)
						return;

					document.autoSubmit.b.value = "Continue (" + countdown + ")";
					countdown--;

					setTimeout("doAutoSubmit();", 1000);
				}
			// ]]></script>';

	show_footer();
	exit;
}

function removeAllAttachments()
{
	global $to_prefix;

	$result = convert_query("
		SELECT value
		FROM {$to_prefix}settings
		WHERE variable = 'attachmentUploadDir'
		LIMIT 1");
	list ($attachmentUploadDir) = mysql_fetch_row($result);
	mysql_free_result($result);

	// !!! This should probably be done in chunks too.
	$result = convert_query("
		SELECT ID_ATTACH, filename
		FROM {$to_prefix}attachments");
	while ($row = mysql_fetch_assoc($result))
	{
		// We're duplicating this from below because it's slightly different for getting current ones.
		$clean_name = strtr($row['filename'], '', 'SZszYAAAAAACEEEEIIIINOOOOOOUUUUYaaaaaaceeeeiiiinoooooouuuuyy');
		$clean_name = strtr($clean_name, array('' => 'TH', '' => 'th', '' => 'DH', '' => 'dh', '' => 'ss', '' => 'OE', '' => 'oe', '' => 'AE', '' => 'ae', '' => 'u'));
		$clean_name = preg_replace(array('/\s/', '/[^\w_\.\-]/'), array('_', ''), $clean_name);
		$enc_name = $row['ID_ATTACH'] . '_' . strtr($clean_name, '.', '_') . md5($clean_name);
		$clean_name = preg_replace('~\.[\.]+~', '.', $clean_name);

		if (file_exists($attachmentUploadDir . '/' . $enc_name))
			$filename = $attachmentUploadDir . '/' . $enc_name;
		else
			$filename = $attachmentUploadDir . '/' . $clean_name;

		@unlink($filename);
	}
	mysql_free_result($result);
}

function getLegacyAttachmentFilename($filename, $attachment_id)
{
	// Remove special accented characters - ie. s (because they won't write to the filesystem well.)
	$clean_name = strtr($filename, '', 'SZszYAAAAAACEEEEIIIINOOOOOOUUUUYaaaaaaceeeeiiiinoooooouuuuyy');
	$clean_name = strtr($clean_name, array('' => 'TH', '' => 'th', '' => 'DH', '' => 'dh', '' => 'ss', '' => 'OE', '' => 'oe', '' => 'AE', '' => 'ae', '' => 'u'));

	// Get rid of dots, spaces, and other weird characters.
	$clean_name = preg_replace(array('/\s/', '/[^\w_\.\-]/'), array('_', ''), $clean_name);

	return $attachment_id . '_' . strtr($clean_name, '.', '_') . md5($clean_name);
}

function addslashes_recursive($var)
{
	if (!is_array($var))
		return addslashes($var);
	else
	{
		foreach ($var as $k => $v)
			$var[$k] = addslashes_recursive($v);
		return $var;
	}
}

// Remove slashes recursively...
function stripslashes_recursive($var, $level = 0)
{
	if (!is_array($var))
		return stripslashes($var);

	// Reindex the array without slashes, this time.
	$new_var = array();

	// Strip the slashes from every element.
	foreach ($var as $k => $v)
		$new_var[stripslashes($k)] = $level > 25 ? null : stripslashes_recursive($v, $level + 1);

	return $new_var;
}

// Update the Settings.php file.
function updateSettingsFile($config_vars)
{
	// Load the file.
	$settingsArray = file($_POST['path_to'] . '/Settings.php');

	if (count($settingsArray) == 1)
		$settingsArray = preg_split('~[\r\n]~', $settingsArray[0]);

	for ($i = 0, $n = count($settingsArray); $i < $n; $i++)
	{
		// Don't trim or bother with it if it's not a variable.
		if (substr($settingsArray[$i], 0, 1) != '$')
			continue;

		$settingsArray[$i] = trim($settingsArray[$i]) . "\n";

		// Look through the variables to set....
		foreach ($config_vars as $var => $val)
		{
			if (strncasecmp($settingsArray[$i], '$' . $var, 1 + strlen($var)) == 0)
			{
				$comment = strstr(substr($settingsArray[$i], strpos($settingsArray[$i], ';')), '#');
				$settingsArray[$i] = '$' . $var . ' = ' . $val . ';' . ($comment == '' ? "\n" : "\t\t" . $comment);

				// This one's been 'used', so to speak.
				unset($config_vars[$var]);
			}
		}

		if (trim(substr($settingsArray[$i], 0, 2)) == '?' . '>')
			$end = $i;
	}

	// This should never happen, but apparently it is happening.
	if (empty($end) || $end < 10)
		$end = count($settingsArray) - 1;

	// Still more?  Add them at the end.
	if (!empty($config_vars))
	{
		$settingsArray[$end++] = '';
		foreach ($config_vars as $var => $val)
			$settingsArray[$end++] = '$' . $var . ' = ' . $val . ';' . "\n";
		$settingsArray[$end] = '?' . '>';
	}

	// Sanity error checking: the file needs to be at least 10 lines.
	if (count($settingsArray) < 10)
		return;

	// Blank out the file - done to fix a oddity with some servers.
	$fp = fopen($_POST['path_to'] . '/Settings.php', 'w');
	fclose($fp);

	// Now actually write.
	$fp = fopen($_POST['path_to'] . '/Settings.php', 'r+');
	$lines = count($settingsArray);
	for ($i = 0; $i < $lines - 1; $i++)
		fwrite($fp, strtr($settingsArray[$i], "\r", ''));

	// The last line should have no \n.
	fwrite($fp, rtrim($settingsArray[$i]));
	fclose($fp);
}

function convert_query($string, $return_error = false)
{
	global $convert_dbs, $to_prefix, $command_line;

	// Debugging?
	if (isset($_REQUEST['debug']))
		$_SESSION['convert_debug'] = !empty($_REQUEST['debug']);

	if (trim($string) == 'TRUNCATE ' . $GLOBALS['to_prefix'] . 'attachments')
		removeAllAttachments();

	if ($convert_dbs != 'mysql')
	{
		$clean = '';
		$old_pos = 0;
		$pos = -1;
		while (true)
		{
			$pos = strpos($string, '\'', $pos + 1);
			if ($pos === false)
				break;
			$clean .= substr($string, $old_pos, $pos - $old_pos);

			while (true)
			{
				$pos1 = strpos($string, '\'', $pos + 1);
				$pos2 = strpos($string, '\\', $pos + 1);
				if ($pos1 === false)
					break;
				elseif ($pos2 == false || $pos2 > $pos1)
				{
					$pos = $pos1;
					break;
				}

				$pos = $pos2 + 1;
			}
			$clean .= '%s';

			$old_pos = $pos + 1;
		}
		$clean .= substr($string, $old_pos);
		$clean = trim(preg_replace('~\s+~s', ' ', $clean));

		if (strpos($string, $to_prefix) === false)
		{
			preg_match('~limit (\d+)(?:, (\d+))?\s*$~is', $string, $limit);
			if (!empty($limit))
			{
				$string = preg_replace('~limit (\d+)(?:, (\d+))?$~is', '', $string);
				if (!isset($limit[2]))
				{
					$limit[2] = $limit[1];
					$limit[1] = 0;
				}
			}

			if ($convert_dbs == 'odbc')
			{
				if (!empty($limit))
					$string = preg_replace('~^\s*select~is', 'SELECT TOP ' . ($limit[1] + $limit[2]), $string);

				$result = @odbc_exec($GLOBALS['odbc_connection'], $string);

				if (!empty($limit) && !empty($limit[1]))
					odbc_fetch_row($result, $limit[1]);
			}
			elseif ($convert_dbs == 'ado')
			{
				if (!empty($limit))
					$string = preg_replace('~^\s*select~is', 'SELECT TOP ' . ($limit[1] + $limit[2]), $string);

				if (PHP_VERSION >= 5)
					eval('
						try
						{
							$result = $GLOBALS[\'ado_connection\']->Execute($string);
						}
						catch (com_exception $err)
						{
							$result = false;
						}');
				else
					$result = @$GLOBALS['ado_connection']->Execute($string);

				if ($result !== false && !empty($limit) && !empty($limit[1]))
					$result->Move($limit[1], 1);
			}
			$not_mysql = true;
		}
		else
			$result = @mysql_query($string);
	}
	else
		$result = @mysql_query($string);

	if ($result !== false || $return_error)
		return $result;
	
	if (empty($not_mysql))
	{
		$mysql_error = mysql_error();
		$mysql_errno = mysql_errno();

		// FIrst log it.
		// In the future this may return the actual file it came from.
		if (isset($_GET['debug']) || isset($_POST['debug']))
			convert_error_hanlder(mysql_error(), $string, 'Converter File', mysql_errno(), true);	

		// Error numbers:
		//    1016: Can't open file '....MYI'
		//    2013: Lost connection to server during query.

		if ($mysql_errno == 1016)
		{
			if (preg_match('~(?:\'([^\.\']+)~', $mysql_error, $match) != 0 && !empty($match[1]))
				mysql_query("
					REPAIR TABLE $match[1]");

			$result = mysql_query($string);

			if ($result !== false)
				return $result;
		}
		elseif ($mysql_errno == 2013)
		{
			$result = mysql_query($string);

			if ($result !== false)
				return $result;
		}
	}
	elseif ($convert_dbs == 'odbc')
		$mysql_error = odbc_errormsg($GLOBALS['odbc_connection']);
	elseif ($convert_dbs == 'ado')
	{
		$error = $GLOBALS['ado_connection']->Errors[0];
		$mysql_error = $error->Description;
	}

	// Get the query string so we pass everything.
	if (isset($_REQUEST['start']))
		$_GET['start'] = $_REQUEST['start'];
	$query_string = '';
	foreach ($_GET as $k => $v)
		$query_string .= '&' . $k . '=' . $v;
	if (strlen($query_string) != 0)
		$query_string = '?' . strtr(substr($query_string, 1), array('&' => '&amp;'));

	if ($command_line)
	{
		print_line("Unsuccessful!  MySQL error message:\n" . mysql_error() . "\n");
		die;
	}

	echo '
			<b>Unsuccessful!</b><br />

			This query:<blockquote>' . nl2br(htmlspecialchars(trim($string))) . ';</blockquote>

			Caused the error:<br />
			<blockquote>' . nl2br(htmlspecialchars($mysql_error)) . '</blockquote>

			<form action="', $_SERVER['PHP_SELF'], $query_string, '" method="post">
				<input type="submit" value="Try again" />
			</form>
		</div>';

	show_footer();
	die;
}

function convert_free_result($result)
{
	// ADO?
	if (!is_resource($result) && is_object($result))
	{
		$result->Close();
		return;
	}

	$type = get_resource_type($result);
	if ($type == 'mysql result')
		mysql_free_result($result);
	elseif ($type == 'odbc result')
		odbc_free_result($result);
}

function convert_fetch_assoc($result)
{
	// Okay, the hardest is ADO (Windows only, by the way.)
	if (!is_resource($result) && is_object($result))
	{
		if ($result->EOF)
			return false;

		$row = array();
		$fields = $result->Fields;
		for ($i = 0, $n = $fields->Count; $i < $n; $i++)
		{
			$field = $fields[$i];
			$row[$field->Name] = $field->Value;
		}
		$result->MoveNext();

		return $row;
	}

	$type = get_resource_type($result);
	if ($type == 'mysql result')
		return mysql_fetch_assoc($result);
	elseif ($type == 'odbc result')
		return odbc_fetch_array($result);
}

function convert_fetch_row($result)
{
	if (!is_resource($result) && is_object($result))
	{
		if ($result->EOF)
			return false;

		$row = array();
		$fields = $result->Fields;
		for ($i = 0, $n = $fields->Count; $i < $n; $i++)
			$row[] = $fields[$i]->Value;
		$result->MoveNext();

		return $row;
	}

	$type = get_resource_type($result);
	if ($type == 'mysql result')
		return mysql_fetch_row($result);
	elseif ($type == 'odbc_result')
		return array_values(odbc_fetch_array($result));
}

function convert_num_rows($result)
{
	if (!is_resource($result) && is_object($result))
	{
		$fields = $result->Fields;
		return $fields->Count;
	}

	$type = get_resource_type($result);
	if ($type == 'mysql result')
		return mysql_num_rows($result);
	elseif ($type == 'odbc result')
		return odbc_num_rows($result);
}

function alterTable($tableName, $knownKeys = '', $knownColumns = '', $alterColumns = '', $reverseKeys = false, $reverseColumns = false, $return_error = false)
{
	global $to_prefix;

	// Shorten this up
	$to_table = $to_prefix . $tableName;

	// Get the keys
	if (!empty($knownKeys))
	{
		$request = convert_query("
			SHOW KEYS
			FROM $to_table");

		$availableKeys = array();
		while ($row = mysql_fetch_assoc($request))
			$availableKeys[] = $row['Key_name'];

		// Flip the keys.
		array_flip($availableKeys);
	}
	else
		$knownKeys = array();

	// Are we dealing with columns also?
	if (!empty($knownColumns))
	{
		$request = convert_query("
			SHOW COLUMNS
			FROM $to_table");

		$availableColumns = array();
		while ($row = mysql_fetch_assoc($request))
			$availableColumns[] = $row['Field'];

		array_flip($availableColumns);
	}
	else
		$knownColumns = array();

	// Column to alter
	if (!empty($alterColumns) && is_array($alterColumns))
		$alterColumns = $alterColumns;
	else
		$alterColumns = array();

	// Check indexes
	foreach ($knownKeys as $key => $value)
	{
		// If we are dropping keys then it should unset the known keys if it's NOT available
		if ($reverseKeys == false && !in_array($key, $availableKeys))
			unset($knownKeys[$key], $knownKeys[$key]);
		// Since we are in reverse and we are adding then unknown the known keys that are available
		elseif ($reverseKeys == true && in_array($key, $availableKeys))
			unset($knownKeys[$key], $knownKeys[$key]);
	}

	// Check columns
	foreach ($knownColumns as $column => $value)
	{
		// Here we reverse things.  If the column is not in then we must add it.
		if ($reverseColumns == false && in_array($column, $availableColumns))
			unset($knownColumns[$column], $knownColumns[$column]);
		// If it's in then we must unset it.
		elseif ($reverseColumns == true && !in_array($column, $availableColumns))
			unset($knownColumns[$column], $knownColumns[$column]);
	}

	// Now merge the three
	$alter = array_merge($alterColumns, $knownKeys, $knownColumns);

	// Now lets see what we want to do with them
	$clause = '';
	foreach ($alter as $key)
		$clause .= "
		$key,";

	// Lets do some altering
	convert_query("
		ALTER TABLE $to_table" .
		substr($clause, 0, -1), $return_error);
}

function copy_smileys($source, $dest)
{
	if (!is_dir($source) || !($dir = opendir($source)))
		return;

	while ($file = readdir($dir))
	{
		if ($file == '.' || $file == '..')
			continue;

		// If we have a directory create it on the destination and copy contents into it!
		if (is_dir($source . '/' . $file))
		{
			if (!is_dir($dest))
				@mkdir($dest . '/' . $file, 0777);
			copy_dir($source . '/' . $file, $dest . '/' . $file);
		}
		else
		{
			if (!is_dir($dest))
				@mkdir($dest . '/' . $file, 0777);
			copy($source . '/' . $file, $dest . '/' . $file);
		}
	}
	closedir($dir);
}

function copy_dir($source, $dest)
{
	if (!is_dir($source) || !($dir = opendir($source)))
		return;

	while ($file = readdir($dir))
	{
		if ($file == '.' || $file == '..')
			continue;

		// If we have a directory create it on the destination and copy contents into it!
		if (is_dir($source . '/'. $file))
		{
			if (!is_dir($dest))
				@mkdir($dest, 0777);
			copy_dir($source . '/' . $file, $dest . '/' . $file);
		}
		else
		{
			if (!is_dir($dest))
				@mkdir($dest, 0777);
			copy($source . '/' . $file, $dest . '/' . $file);
		}
	}
	closedir($dir);
}

// Removes special entities from strings.  Compatibility...
function un_htmlspecialchars($string)
{
	return strtr($string, array_flip(get_html_translation_table(HTML_SPECIALCHARS, ENT_QUOTES)) + array('&#039;' => '\'', '&nbsp;' => ' '));
}

// CLI
function cmdStep0()
{
	global $time_start;
	$time_start = time();

	ob_end_clean();
	ob_implicit_flush(true);
	@set_time_limit(0);

	if (!isset($_SERVER['argv']))
		$_SERVER['argv'] = array();

	// If its empty, force help.
	if (empty($_SERVER['argv'][1]))
		$_SERVER['argv'][1] = '--help';

	// Lets get the path_to and path_from
	foreach ($_SERVER['argv'] as $i => $arg)
	{
		// Trim spaces.
		$arg = trim($arg);

		if (preg_match('~^--path_to=(.+)$~', $arg, $match) != 0)
			$_POST['path_to'] = substr($match[1], -1) == '/' ? substr($match[1], 0, -1) : $match[1];
		elseif (preg_match('~^--path_from=(.+)$~', $arg, $match) != 0)
			$_POST['path_from'] = substr($match[1], -1) == '/' ? substr($match[1], 0, -1) : $match[1];
		elseif (preg_match('~^--db_pass=(.+)?$~', $arg, $match) != 0)
			$_POST['db_pass'] = isset($match[1]) ? $match[1] : '';
		elseif ($arg == '--debug')
			$_GET['debug'] = 1;
		elseif (preg_match('~^--convert_script=(.+)$~', $arg, $match) != 0)
			$_REQUEST['convert_script'] = $match[1];
		elseif ($arg == '--help' || $i = 0)
		{
			echo 'SMF Command-line Converter
Usage: /path/to/php -f ' . basename(__FILE__) . ' -- [OPTION]...

    --path_to               Path to SMF (' . dirname(__FILE__) . ').
    --path_from             Path to the software that you are converting from.
    --convert_script        The name of the script. (old_forum_to_smf.sql)
    --db_pass               SMF database password. "The MySQL password (for verification only.)"
    --debug                 Output debugging information.';
			echo "\n";
			exit;
		}

		// We have extra params.
		if (preg_match('~^--(.+)=(.+)$~', $arg, $match) != 0 && !array_key_exists($match[1], $_POST))
			$_POST[$match[1]] = $match[2];
	}

	// Do we have the paths and passwords?
	if (!isset($_POST['path_to']))
		print_error('ERROR: You must enter the path_to in order to convert.', true);
	elseif (!isset($_POST['path_from']))
		print_error('ERROR: You must enter the path_from in order to convert.', true);
	elseif (!isset($_POST['db_pass']))
		print_error('ERROR: You must enter SMF\'s database password in order to convert.', true);
	else
	{
		$_GET['step'] = 1;
		initialize_inputs();
		doStep1();
	}

	exit;
}

function print_error($message, $fatal = false)
{
	static $fp = null;

	if ($fp === null)
		$fp = fopen('php://stderr', 'wb');

	fwrite($fp, $message);

	if ($fatal)
		exit;
}

function print_line($line)
{
	global $command_line;

	if ($command_line)
		print_error($line);
	else
		echo $line;
}

// Get the ID_MEMBER associated with the specified message.
function getMsgMemberID($messageID)
{
	global $to_prefix;

	// Find the topic and make sure the member still exists.
	$result = convert_query("
		SELECT IFNULL(mem.ID_MEMBER, 0)
		FROM {$to_prefix}messages AS m
			LEFT JOIN {$to_prefix}members AS mem ON (mem.ID_MEMBER = m.ID_MEMBER)
		WHERE m.ID_MSG = " . (int) $messageID . "
		LIMIT 1");
	if (mysql_num_rows($result) > 0)
		list ($memberID) = mysql_fetch_row($result);
	// The message doesn't even exist.
	else
		$memberID = 0;
	mysql_free_result($result);

	return $memberID;
}

// Handles our errors.
function convert_error_hanlder($error_level, $error_string, $file, $line, $is_database_error = false)
{
	global $command_line;

	// Our error_log
	$convert_error_log = dirname(__FILE__) . '/convert_error_log';

	// Is it Database Specific?
	if (!empty($is_database_error))
	{
		/*
			Note, Unlike php this is slightly different.
			error_level = errror from database
			error_string = original query
		*/

		// The array is easier than using \r as well makes it easier for command line.
		$error_array = array(
			"",
			"The database encountered an error on line (from query), " . $line . ".", // . ", from file, " . $file . ".";
			"The error received was:",
			"---",
			$error_level,
			"---",
			"The query ran was:",
			"---",
			$error_string,
			"---",
		);
	}
	else
	{
		// Generate a simple error message familer to PHP errors, expect we do them one better.
		$error_array = array(
			'',
			$error_level % 255 == E_ERROR ? 'Error' : ($error_level % 255 == E_WARNING ? 'Warning' : 'Notice') . ': ' . $error_string . ' in ' . $file . ' on line ' . $line,
			"Backtrace report",
			"---"
		);

		// Lets leave a paper trail.
		$backtrace = debug_backtrace();

		// Now loop through our backtrace for output.
		foreach ($backtrace as $trail)
		{
			foreach ($trail as $key => $value)
			{
				if (!is_array($value))
					$error_array[] = "\t " . $key . " (key): " . $value;
				else
				{
					$error_array[] = "\t" . $key . " (key): ";

					foreach ($value as $vkey => $vvalue)
						$error_array[] = "\t\t" . $vkey . " (key): " . $vvalue;
				}
			}

			// Add an extra return.
			$error_array[] = "";
		}
	}

	// Open the file up for writing (with pointer at end).
	$fp = fopen($convert_error_log, 'a+');

	// Can't open it!
	if (!$fp)
		$no_write_file = true;

	// Send it out to the error log. Only if we can though.
	if (empty($no_write_file))
	{
		// The world implodes!
		$error_data = implode("\r", $error_array);

		fwrite($fp, $error_data);
		fclose($fp);
	}	

	// Based on if we are command line or not, this will handle things for us.
	if ($command_line && !$is_database_error)
		print_error($error_data[1]);
	// If its not command lind and not a database error, echo the error info (without backtrack).
	elseif (!$is_database_error)
		echo $error_array[1];
}
?>